@@ -117,6 +117,13 @@ static gint       xfce_appfinder_window_sort_items                    (GtkTreeMo
                                                                        GtkTreeIter                 *b,
                                                                        gpointer                     data);
 static gboolean   xfce_appfinder_should_sort_icon_view                (void);
+static void       xfce_appfinder_window_update_frecency               (XfceAppfinderWindow         *window,
+                                                                       GtkTreeModel                *model,
+                                                                       const gchar                 *uri);
+static gint       xfce_appfinder_window_sort_items_frecency           (GtkTreeModel                *model,
+                                                                       GtkTreeIter                 *a,
+                                                                       GtkTreeIter                 *b,
+                                                                       gpointer                     data);
 
 struct _XfceAppfinderWindowClass
 {
@@ -153,7 +160,6 @@ struct _XfceAppfinderWindow
 
   GtkWidget                  *bbox;
   GtkWidget                  *button_launch;
-  GtkWidget                  *button_preferences;
 
   GarconMenuDirectory        *filter_category;
   gchar                      *filter_text;
@@ -219,7 +225,7 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   xfconf_g_property_bind (window->channel, "/category-icon-size", G_TYPE_UINT,
                           G_OBJECT (window->category_model), "icon-size");
 
-  window->model = xfce_appfinder_model_get ();
+  window->model = xfce_appfinder_model_get (xfconf_channel_get_bool (window->channel, "/sort-by-frecency", FALSE));
   xfconf_g_property_bind (window->channel, "/item-icon-size", G_TYPE_UINT,
                           G_OBJECT (window->model), "icon-size");
 
@@ -276,10 +282,12 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   gtk_entry_completion_set_model (completion, GTK_TREE_MODEL (window->model));
   gtk_entry_completion_set_match_func (completion, xfce_appfinder_window_completion_match_func, window, NULL);
   gtk_entry_completion_set_text_column (completion, XFCE_APPFINDER_MODEL_COLUMN_COMMAND);
-  gtk_entry_completion_set_popup_completion (completion, TRUE);
   gtk_entry_completion_set_popup_single_match (completion, TRUE);
   gtk_entry_completion_set_inline_completion (completion, TRUE);
 
+  if (xfconf_channel_get_bool (window->channel, "/disable-completion", FALSE))
+    gtk_entry_completion_set_popup_completion (completion, FALSE);
+
   window->paned = pane = gtk_paned_new (GTK_ORIENTATION_HORIZONTAL);
   gtk_box_pack_start (GTK_BOX (vbox), pane, TRUE, TRUE, 0);
   integer = xfconf_channel_get_int (window->channel, "/last/pane-position", DEFAULT_PANED_POSITION);
@@ -340,12 +348,13 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 0);
   gtk_widget_show (hbox);
 
-  window->button_preferences = button = gtk_button_new_with_mnemonic (_("_Preferences"));
+  button = gtk_button_new_with_mnemonic (_("_Preferences"));
   gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
   g_signal_connect (G_OBJECT (button), "clicked",
       G_CALLBACK (xfce_appfinder_window_preferences), window);
   gtk_widget_set_valign (button, GTK_ALIGN_CENTER);
   gtk_button_set_always_show_image(GTK_BUTTON(button), TRUE);
+  gtk_widget_show (button);
 
   image = gtk_image_new_from_icon_name (XFCE_APPFINDER_STOCK_PREFERENCES, GTK_ICON_SIZE_BUTTON);
   gtk_button_set_image (GTK_BUTTON (button), image);
@@ -700,7 +709,10 @@ xfce_appfinder_window_view (XfceAppfinderWindow *window)
   gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (window->filter_model), xfce_appfinder_window_item_visible, window, NULL);
 
   window->sort_model = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (window->filter_model));
-  gtk_tree_sortable_set_default_sort_func (GTK_TREE_SORTABLE (window->sort_model), xfce_appfinder_window_sort_items, window->entry, NULL);
+  if (xfconf_channel_get_bool (window->channel, "/sort-by-frecency", FALSE))
+    gtk_tree_sortable_set_default_sort_func (GTK_TREE_SORTABLE (window->sort_model), xfce_appfinder_window_sort_items_frecency, window->entry, NULL);
+  else
+    gtk_tree_sortable_set_default_sort_func (GTK_TREE_SORTABLE (window->sort_model), xfce_appfinder_window_sort_items, window->entry, NULL);
 
   if (icon_view)
     {
@@ -1266,10 +1278,6 @@ xfce_appfinder_window_entry_changed_idle (gpointer data)
     {
       gtk_widget_set_sensitive (window->button_launch, IS_STRING (text));
 
-      /* Clear entry's error icon and tooltip */
-      gtk_entry_set_icon_from_icon_name (GTK_ENTRY (window->entry), GTK_ENTRY_ICON_PRIMARY, NULL);
-      gtk_entry_set_icon_tooltip_text (GTK_ENTRY (window->entry), GTK_ENTRY_ICON_PRIMARY, NULL);
-
       pixbuf = xfce_appfinder_model_get_icon_for_command (window->model, text);
       xfce_appfinder_window_update_image (window, pixbuf);
       if (pixbuf != NULL)
@@ -1675,17 +1683,13 @@ xfce_appfinder_window_item_visible (GtkTreeModel *model,
                                     gpointer      data)
 {
   XfceAppfinderWindow *window = XFCE_APPFINDER_WINDOW (data);
-  /* don't use the casefolded key generated by gtk */
   const gchar *filter_string = gtk_entry_get_text (GTK_ENTRY (window->entry));
-  gchar *filter_string_down = g_ascii_strdown (filter_string, -1);
 
-  gboolean res = xfce_appfinder_model_get_visible (XFCE_APPFINDER_MODEL (model), iter,
+  return xfce_appfinder_model_get_visible (XFCE_APPFINDER_MODEL (model),
+                                           iter,
                                            window->filter_category,
-                                           filter_string_down);
-
-  g_free(filter_string_down);
-
-  return res;
+                                           filter_string,
+                                           window->filter_text);
 }
 
 
@@ -1823,6 +1827,7 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window,
   gboolean      regular_command = FALSE;
   gboolean      save_cmd;
   gboolean      only_custom_cmd = FALSE;
+  const gchar  *uri;
 
   screen = gtk_window_get_screen (GTK_WINDOW (window));
   if (gtk_widget_get_visible (window->paned))
@@ -1836,6 +1841,7 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window,
       if (xfce_appfinder_window_view_get_selected (window, &model, &iter))
         {
           child_model = model;
+          gtk_tree_model_get (model, &iter, XFCE_APPFINDER_MODEL_COLUMN_URI, &uri, -1);
 
           if (GTK_IS_TREE_MODEL_SORT (model) || xfce_appfinder_should_sort_icon_view ())
             {
@@ -1847,6 +1853,7 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window,
           gtk_tree_model_filter_convert_iter_to_child_iter (GTK_TREE_MODEL_FILTER (child_model), &child_iter, &iter);
           result = xfce_appfinder_model_execute (window->model, &child_iter, screen, &regular_command, &error);
 
+          xfce_appfinder_window_update_frecency (window, model, uri);
           if (!result && regular_command)
             {
               gtk_tree_model_get (model, &child_iter, XFCE_APPFINDER_MODEL_COLUMN_COMMAND, &cmd, -1);
@@ -1894,6 +1901,32 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window,
 
 
 
+static void
+xfce_appfinder_window_update_frecency (XfceAppfinderWindow *window,
+                                        GtkTreeModel        *model,
+                                        const gchar         *uri)
+{
+  GFile      *gfile;
+  gchar      *desktop_id;
+  GError     *error = NULL;
+
+  if (uri != NULL)
+    {
+      gfile = g_file_new_for_uri (uri);
+      desktop_id = g_file_get_basename (gfile);
+      g_object_unref (G_OBJECT (gfile));
+      APPFINDER_DEBUG ("desktop : %s", desktop_id);
+
+      model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (window->filter_model));
+      xfce_appfinder_model_update_frecency (XFCE_APPFINDER_MODEL (model), desktop_id, &error);
+
+      g_free (desktop_id);
+      g_free (error);
+    }
+}
+
+
+
 void
 xfce_appfinder_window_set_expanded (XfceAppfinderWindow *window,
                                     gboolean             expanded)
@@ -1921,9 +1954,6 @@ xfce_appfinder_window_set_expanded (XfceAppfinderWindow *window,
       gtk_window_set_geometry_hints (GTK_WINDOW (window), NULL, &hints, GDK_HINT_MAX_SIZE);
     }
 
-  /* show/hide preferences button */
-  gtk_widget_set_visible (window->button_preferences, expanded);
-
   /* show/hide pane with treeviews */
   gtk_widget_set_visible (window->paned, expanded);
 
@@ -1965,13 +1995,11 @@ xfce_appfinder_window_sort_items (GtkTreeModel *model,
 
   gtk_tree_model_get (model, a, XFCE_APPFINDER_MODEL_COLUMN_TITLE, &title_a, -1);
   normalized = g_utf8_normalize (title_a, -1, G_NORMALIZE_ALL);
-  g_free (title_a);
   title_a = g_utf8_casefold (normalized, -1);
   g_free (normalized);
 
   gtk_tree_model_get (model, b, XFCE_APPFINDER_MODEL_COLUMN_TITLE, &title_b, -1);
   normalized = g_utf8_normalize (title_b, -1, G_NORMALIZE_ALL);
-  g_free (title_b);
   title_b = g_utf8_casefold (normalized, -1);
   g_free (normalized);
 
@@ -1996,6 +2024,35 @@ xfce_appfinder_window_sort_items (GtkTreeModel *model,
 
 
 
+static gint
+xfce_appfinder_window_sort_items_frecency  (GtkTreeModel *model,
+                                            GtkTreeIter  *a,
+                                            GtkTreeIter  *b,
+                                            gpointer      data)
+{
+  guint        a_freq, b_freq;
+  guint64      a_rec, b_rec;
+  guint        a_res, b_res;
+
+  gtk_tree_model_get (model, a, XFCE_APPFINDER_MODEL_COLUMN_FREQUENCY, &a_freq, -1);
+  gtk_tree_model_get (model, b, XFCE_APPFINDER_MODEL_COLUMN_FREQUENCY, &b_freq, -1);
+
+  gtk_tree_model_get (model, a, XFCE_APPFINDER_MODEL_COLUMN_RECENCY, &a_rec, -1);
+  gtk_tree_model_get (model, b, XFCE_APPFINDER_MODEL_COLUMN_RECENCY, &b_rec, -1);
+
+  a_res = xfce_appfinder_model_calculate_frecency (a_freq, a_rec);
+  b_res = xfce_appfinder_model_calculate_frecency (b_freq, b_rec);
+
+  /* Sort according the frecency */
+  if (b_res - a_res != 0)
+    return b_res - a_res;
+
+  /* If they have the same frecency, fallback to the alphabetical order */
+  return xfce_appfinder_window_sort_items (model, a, b, data);
+}
+
+
+
 /*
  * Checks for gtk => 3.22.27 during runtime.
  * This is necessary because sort model for icon view did not work as expected.
